part of tecfy_basic_package;

abstract class IApiService {
  static Function(dynamic response)? unauthrizedCallback;
  Future<OperationResult<Map>> httpGet(String url);
  Future<OperationResult<dynamic>> httpPost(String url, dynamic data,
      [bool asString = false]);
  Future<OperationResult<dynamic>> httpPut(String url, Map data);
  Future<OperationResult<dynamic>> httpPostEx(String url, dynamic data,
      [bool addToken = true]);
  Future<OperationResult<dynamic>> httpPutEx(String url, dynamic data);
  Future<OperationResult<dynamic>> httpGetDynamic(String url);
  Future<OperationResult<dynamic>> httpDeleteDynamic(String url);
  Future<OperationResult<dynamic>> httpDeleteDynamicEx(String url);
  Future<OperationResult<String>> httpGetString(String url);
  Future<OperationResult<String>> httpPostFile(
      String url, Uint8List fileData, String name,
      [bool isImage = true]);
}

class ApiService implements IApiService {
  /// backend token
  static String get token => ConfigService.token;

  static Map<String, String>? _baseHeader;
  Map<String, String> get baseHeader {
    if (_baseHeader == null) {
      _baseHeader = {
        'AppVersion': TecfyBasicApp.appVersion,
        'DeviceType': kIsWeb
            ? 'WebBrowser'
            : Platform.isAndroid
                ? 'Android'
                : Platform.isIOS
                    ? 'iOS'
                    : Platform.isWindows
                        ? 'Windows'
                        : Platform.isMacOS
                            ? 'MacOS'
                            : Platform.isLinux
                                ? 'Linux'
                                : Platform.isFuchsia
                                    ? 'Fuchsia'
                                    : 'Unknown',
      };
      if (!kIsWeb) {
        _baseHeader!['Accept-Encoding'] = 'gzip';
      }
    }
    return _baseHeader!;
  }

  /// call backend and return OperationResult with Map
  Future<OperationResult<Map>> httpGet(String url) async {
    try {
      var result = await httpGetString(url);
      if (!result.success)
        return OperationResult(
            success: false,
            message: result.message,
            innerError: result.innerError);
      var map = json.decode(result.data ?? '');
      if (kDebugMode) print('$url: $map');

      return OperationResult(success: true, data: map);
    } catch (err, t) {
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  /// custom Encodeer for DateTime issue
  static dynamic customEncode(dynamic item) {
    if (item is DateTime) {
      return intl.DateFormat("yyyy-MM-ddTHH:mm:ss").format(item);
    }
    return item;
  }

  /// call backend and return OperationResult with Map
  Future<OperationResult<dynamic>> httpPost(String url, dynamic data,
      [bool asString = false]) async {
    try {
      var httpClient = http.Client();
      if (kDebugMode) print('POST: ${await _getFullUrl(url)}');

      if (asString)
        data =
            json.encode(data, toEncodable: customEncode).replaceAll('\"', '\'');
      var body = json.encode(data, toEncodable: customEncode);

      var headers = {'content-type': 'application/json'};
      for (var h in baseHeader.keys) headers[h] = baseHeader[h]!;
      if (token.isNotEmpty) headers['Authorization'] = 'Bearer $token';

      http.Response response = await httpClient.post(
        await _getFullUrl(url),
        headers: headers,
        body: utf8.encode(body),
      );
      if (response.statusCode != 200) //&& response.statusCode !=307
      {
        print('statusCode: ${response.statusCode}');
        return OperationResult(
            success: false, message: 'Result code = ${response.statusCode}');
      }
      var r = jsonDecode(response.body);
      httpClient.close();
      return OperationResult(
          success: r['success'],
          innerError: r['innerError'],
          minVersion: r['minVersion'],
          lastVersion: r['lastVersion'],
          message: r['messageAr'] ?? r['errorString'],
          data: r['data'],
          errorCodeString: r['errorCodeString']);
      // } else
      //   return OperationResult(success: false, message: 'Result code = ${result.statusCode}');
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  /// call backend and return OperationResult with Map
  Future<OperationResult<dynamic>> httpPut(String url, Map data) async {
    try {
      var httpClient = http.Client();

      if (kDebugMode) print('PUT: ${await _getFullUrl(url)}');

      var headers = {'content-type': 'application/json'};
      for (var h in baseHeader.keys) headers[h] = baseHeader[h]!;
      if (token.isNotEmpty) headers['Authorization'] = 'Bearer $token';

      http.Response response = await httpClient.put(
        await _getFullUrl(url),
        headers: headers,
        body: json.encode(data, toEncodable: customEncode),
      );

      if (response.statusCode != 200) //&& response.statusCode !=307
      {
        print(response.statusCode);
        return OperationResult(
            success: false, message: 'Result code = ${response.statusCode}');
      }

      String reply = response.body;

      httpClient.close();

      var r = jsonDecode(reply);
      return OperationResult(
        success: r['success'],
        innerError: r['innerError'],
        minVersion: r['minVersion'],
        lastVersion: r['lastVersion'],
        message: r['messageAr'] ?? r['errorString'],
        data: r['data'],
        errorCodeString: r['errorCodeString'],
      );
      // } else
      //   return OperationResult(success: false, message: 'Result code = ${result.statusCode}');
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  /// call backend and return OperationResult with Map
  Future<OperationResult<dynamic>> httpPostEx(String url, dynamic data,
      [bool addToken = true]) async {
    try {
      if (kDebugMode) print('POST: ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var result = await httpClient.post(await _getFullUrl(url),
          body: json.encode(data, toEncodable: customEncode),
          headers: addToken
              ? {
                  'content-type': 'application/json',
                  ...baseHeader,
                  'Authorization': 'Bearer $token'
                }
              : {
                  'content-type': 'application/json',
                  ...baseHeader,
                });
      httpClient.close();
      if (result.statusCode != 200) {
        print('XXXXXXX  Response status code ${result.statusCode}');
        return OperationResult(
            success: false, message: 'Result code = ${result.statusCode}');
      }

      String reply = result.body;
      var r = jsonDecode(reply);

      return OperationResult(
          success: r['success'] ?? false,
          innerError: r['innerError'],
          minVersion: r['minVersion'],
          lastVersion: r['lastVersion'],
          message: r['messageAr'] ??
              r['messageEn'] ??
              r['errorString'] ??
              r['fullMessagesString'] ??
              r['errorCodeString'] ??
              'No Response',
          data: r['data'],
          errorCodeString: r['errorCodeString']);
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  /// call backend and return OperationResult with Map
  Future<OperationResult<dynamic>> httpPutEx(String url, dynamic data) async {
    try {
      if (kDebugMode) print('PUT: ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var result = await httpClient.put(await _getFullUrl(url),
          body: json.encode(data, toEncodable: customEncode),
          headers: {
            'content-type': 'application/json',
            ...baseHeader,
            'Authorization': 'Bearer $token'
          });
      httpClient.close();
      if (result.statusCode != 200) //&& response.statusCode !=307
      {
        print(result.statusCode);
        return OperationResult(
            success: false, message: 'Result code = ${result.statusCode}');
      }

      String reply = result.body;

      var r = jsonDecode(reply);
      return OperationResult(
          success: r['success'],
          innerError: r['innerError'],
          minVersion: r['minVersion'],
          lastVersion: r['lastVersion'],
          message:
              r['messageAr'] ?? r['errorString'] ?? r['fullMessagesString'],
          data: r['data'],
          errorCodeString: r['errorCodeString']);
      // } else
      //   return OperationResult(success: false, message: 'Result code = ${result.statusCode}');
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  /// call backend and return OperationResult with Map
  Future<OperationResult<dynamic>> httpGetDynamic(String url) async {
    try {
      if (kDebugMode)
        print('GET --------------------- ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var response = await httpClient.get(await _getFullUrl(url), headers: {
        'content-type': 'application/json',
        ...baseHeader,
        'Authorization': 'Bearer $token'
      });
      httpClient.close();
      if (response.statusCode != 200) {
        print('XXXXXXXXXXX GET status is ${response.statusCode} $url');
        if (response.statusCode == 401 &&
            IApiService.unauthrizedCallback != null) {
          IApiService.unauthrizedCallback?.call(response.statusCode);
        }
        return OperationResult(
            success: false, message: 'Result code = ${response.statusCode}');
      } else if (kDebugMode) print('GET response $url');
      String reply = response.body;

      if (reply.length == 0)
        return OperationResult(success: false, message: 'Result is empty');

      var r = jsonDecode(reply);

      if (r['success'] != true) {
        print(r);
        if (r['errorCodeString'] == 'InvalidAuthentication' &&
            IApiService.unauthrizedCallback != null) {
          IApiService.unauthrizedCallback?.call(r);
        }
      }
      return OperationResult(
        success: r['success'],
        innerError: r['innerError'],
        minVersion: r['minVersion'],
        lastVersion: r['lastVersion'],
        errorCodeString: r['errorCodeString'],
        message: r['messageAr'] ??
            r['errorString'] ??
            r['errorCodeString'] ??
            r['fullMessagesString'] ??
            'No Server Error!',
        data: r['data'],
      );
    } catch (err, t) {
      print('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx ERROR: GET: $url');
      print(err);
      print(t);
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  /// call backend and return OperationResult with Map
  Future<OperationResult<dynamic>> httpDeleteDynamic(String url) async {
    try {
      if (kDebugMode)
        print('DELETE --------------------- ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var response = await httpClient.delete(await _getFullUrl(url), headers: {
        'content-type': 'application/json',
        ...baseHeader,
        'Authorization': 'Bearer $token'
      });
      httpClient.close();
      if (response.statusCode != 200)
        return OperationResult(
            success: false, message: 'Result code = ${response.statusCode}');

      String reply = response.body;

      if (reply.length == 0)
        return OperationResult(success: false, message: 'Result is empty');

      var r = jsonDecode(reply);

      if (r['success'] != true) print(r);
      return OperationResult(
        success: r['success'],
        innerError: r['innerError'],
        minVersion: r['minVersion'],
        lastVersion: r['lastVersion'],
        errorCodeString: r['errorCodeString'],
        message: r['messageAr'] ??
            r['errorString'] ??
            r['errorCodeString'] ??
            r['fullMessagesString'] ??
            'No Server Error!',
        data: r['data'],
      );
    } catch (err, t) {
      print('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx   ERROR: DELETE: $url');
      print(err);
      print(t);
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  /// call backend and return OperationResult with Map
  Future<OperationResult<dynamic>> httpDeleteDynamicEx(String url) async {
    try {
      if (kDebugMode)
        print('DELETE --------------------- ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var headers = {'content-type': 'application/json'};
      for (var h in baseHeader.keys) headers[h] = baseHeader[h]!;
      if (token.isNotEmpty) headers['Authorization'] = 'Bearer $token';

      http.Response response =
          await httpClient.delete(await _getFullUrl(url), headers: headers);

      if (response.statusCode != 200)
        return OperationResult(
            success: false, message: 'Result code = ${response.statusCode}');

      String reply = response.body;
      httpClient.close();

      if (reply.length == 0)
        return OperationResult(success: false, message: 'Result is empty');

      var r = jsonDecode(reply);

      if (r['success'] != true) print(r);
      return OperationResult(
        success: r['success'],
        innerError: r['innerError'],
        minVersion: r['minVersion'],
        lastVersion: r['lastVersion'],
        errorCodeString: r['errorCodeString'],
        message: r['messageAr'] ??
            r['errorString'] ??
            r['errorCodeString'] ??
            r['fullMessagesString'] ??
            'No Server Error!',
        data: r['data'],
      );
    } catch (err, t) {
      print('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx   ERROR: DELETE: $url');
      print(err);
      print(t);
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  Future<OperationResult<String>> httpGetString(String url) async {
    try {
      var httpClient = http.Client();
      var result = await httpClient.get(await _getFullUrl(url), headers: {
        ...baseHeader,
        //'Authorization': 'Bearer $token'
      });
      httpClient.close();
      if (result.statusCode == 200) {
        return OperationResult(success: true, data: result.body);
      } else
        return OperationResult(
            success: false, message: 'Result code = ${result.statusCode}');
    } catch (err, t) {
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  /// static String apiURL = 'http://app.tecfy.co/api/';
  static String? get apiURL => ConfigService.apiURL;
  static Future<Uri> _getFullUrl(String url) async {
    if (url.startsWith('http')) return Uri.parse(url);
    while (apiURL == null || (apiURL?.isEmpty ?? true)) {
      await Future.delayed(Duration(milliseconds: 100));
    }
    return Uri.parse("$apiURL$url");
  }

  /// call backend and post file like images
  Future<OperationResult<String>> httpPostFile(
      String url, Uint8List fileData, String name,
      [bool isImage = true]) async {
    String dataString;
    try {
      //print('POST:---- ${_getFullUrl(url)}   $name');
      //var httpClient = http.Client();
      var request = http.MultipartRequest("POST", await _getFullUrl(url));

      request.headers['Authorization'] = 'Bearer $token';
      // request.fields['user'] = 'someone@somewhere.com';
      request.files.add(http.MultipartFile.fromBytes(
        'file',
        fileData,
        filename: name,
        contentType: isImage
            ? MediaType("image", path.extension(name))
            : MediaType('application', 'x-tar'),
      ));

      request.headers["Authorization"] = "Bearer $token";
      // request.headers["Accept"] = vimeoUploadHeader;
      request.headers["Content-Type"] = "image/jpg";
      //request.headers["Content-Length"] = fileData.length.toString();
      var response = await request.send();
      if (response.statusCode == 200) {
        if (kDebugMode) print("Uploaded!!!");

        var data = (await response.stream.toBytes());
        if (kDebugMode) print(data.length);
        dataString = utf8.decode(data);
        var r = json.decode(dataString);
        return OperationResult(
            success: r['success'],
            innerError: r['innerError'],
            minVersion: r['minVersion'],
            lastVersion: r['lastVersion'],
            message: r['messageAr'] ?? r['errorString'],
            data: r['data'],
            errorCodeString: r['errorCodeString']);
      }
      return OperationResult(
          success: false, message: 'Error HttpPosFile: ${response.statusCode}');
    } catch (err, t) {
      print('XXXXXXXXXXXXXXXX post file error: $err');
      print(t);
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }
}
